from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Optional

if TYPE_CHECKING:
    from .solid import SolidDefinition
    from .resource import ResourceDefinition


class VersionStrategy(ABC):
    """Abstract class for defining a strategy to version solids and resources.

    When subclassing, `get_solid_version` must be implemented, and `get_resource_version` can be
    optionally implemented.

    `get_solid_version` should ingest a SolidDefinition, and `get_resource_version` should ingest a
    ResourceDefinition. From that,  each synthesize a unique string called a `version`, which will
    be tagged to outputs of that solid in the pipeline. Providing a `VersionStrategy` instance to a
    job will enable memoization on that job, such that only steps whose outputs do not have an
    up-to-date version will run.
    """

    @abstractmethod
    def get_solid_version(self, solid_def: "SolidDefinition") -> str:
        pass

    def get_resource_version(
        self, resource_def: "ResourceDefinition"  # pylint: disable=unused-argument
    ) -> Optional[str]:
        return None



from .config import ConfigMapping
from .decorators import (
    asset_sensor,
    composite_solid,
    daily_schedule,
    failure_hook,
    graph,
    hook,
    hourly_schedule,
    lambda_solid,
    monthly_schedule,
    op,
    pipeline,
    repository,
    schedule,
    sensor,
    solid,
    success_hook,
    weekly_schedule,
)
from .dependency import (
    DependencyDefinition,
    MultiDependencyDefinition,
    Node,
    NodeHandle,
    SolidInputHandle,
    SolidInvocation,
    SolidOutputHandle,
)
from .event_metadata import (
    EventMetadata,
    EventMetadataEntry,
    FloatMetadataEntryData,
    IntMetadataEntryData,
    JsonMetadataEntryData,
    MarkdownMetadataEntryData,
    PathMetadataEntryData,
    PythonArtifactMetadataEntryData,
    TextMetadataEntryData,
    UrlMetadataEntryData,
)
from .events import (
    AssetKey,
    AssetMaterialization,
    DynamicOutput,
    ExpectationResult,
    Failure,
    HookExecutionResult,
    Materialization,
    Output,
    RetryRequested,
    TypeCheck,
)
from .executor import (
    ExecutorDefinition,
    ExecutorRequirement,
    default_executors,
    executor,
    in_process_executor,
    multiple_process_executor_requirements,
    multiprocess_executor,
)
from .graph import GraphDefinition
from .hook import HookDefinition
from .input import In, InputDefinition, InputMapping
from .intermediate_storage import IntermediateStorageDefinition, intermediate_storage
from .logger import LoggerDefinition, build_init_logger_context, logger
from .mode import ModeDefinition
from .output import DynamicOut, DynamicOutputDefinition, Out, OutputDefinition, OutputMapping
from .partition import Partition, PartitionScheduleDefinition, PartitionSetDefinition
from .partitioned_schedule import schedule_from_partitions
from .pipeline import PipelineDefinition
from .pipeline_base import IPipeline
from .pipeline_sensor import (
    PipelineFailureSensorContext,
    RunStatusSensorContext,
    RunStatusSensorDefinition,
    pipeline_failure_sensor,
    run_status_sensor,
)
from .preset import PresetDefinition
from .reconstructable import (
    ReconstructablePipeline,
    build_reconstructable_pipeline,
    reconstructable,
)
from .repository import RepositoryData, RepositoryDefinition
from .resource import ResourceDefinition, make_values_resource, resource
from .run_config_schema import RunConfigSchema, create_run_config_schema
from .run_request import JobType, RunRequest, SkipReason
from .schedule import ScheduleDefinition, ScheduleEvaluationContext, ScheduleExecutionContext
from .sensor import (
    AssetSensorDefinition,
    SensorDefinition,
    SensorEvaluationContext,
    SensorExecutionContext,
)
from .solid import CompositeSolidDefinition, NodeDefinition, SolidDefinition
from .solid_container import create_execution_structure
from .time_window_partitions import (
    PartitionedConfig,
    daily_partitioned_config,
    hourly_partitioned_config,
    monthly_partitioned_config,
    weekly_partitioned_config,
)



from dagster.builtins import Any, Bool, Float, Int, Nothing, String
from dagster.config import Enum, EnumValue, Field, Permissive, Selector, Shape
from dagster.config.config_schema import ConfigSchema
from dagster.config.config_type import Array, Noneable, ScalarUnion
from dagster.core.definitions import (
    AssetKey,
    AssetMaterialization,
    AssetSensorDefinition,
    CompositeSolidDefinition,
    ConfigMapping,
    DependencyDefinition,
    DynamicOut,
    DynamicOutput,
    DynamicOutputDefinition,
    EventMetadata,
    EventMetadataEntry,
    ExecutorDefinition,
    ExecutorRequirement,
    ExpectationResult,
    Failure,
    FloatMetadataEntryData,
    GraphDefinition,
    HookDefinition,
    In,
    InputDefinition,
    InputMapping,
    IntMetadataEntryData,
    IntermediateStorageDefinition,
    JsonMetadataEntryData,
    LoggerDefinition,
    MarkdownMetadataEntryData,
    Materialization,
    ModeDefinition,
    MultiDependencyDefinition,
    Out,
    Output,
    OutputDefinition,
    OutputMapping,
    Partition,
    PartitionScheduleDefinition,
    PartitionSetDefinition,
    PathMetadataEntryData,
    PipelineDefinition,
    PipelineFailureSensorContext,
    PresetDefinition,
    PythonArtifactMetadataEntryData,
    RepositoryData,
    RepositoryDefinition,
    ResourceDefinition,
    RetryRequested,
    RunRequest,
    RunStatusSensorContext,
    RunStatusSensorDefinition,
    ScheduleDefinition,
    ScheduleEvaluationContext,
    ScheduleExecutionContext,
    SensorDefinition,
    SensorEvaluationContext,
    SensorExecutionContext,
    SkipReason,
    SolidDefinition,
    SolidInvocation,
    TextMetadataEntryData,
    TypeCheck,
    UrlMetadataEntryData,
    asset_sensor,
    build_init_logger_context,
    composite_solid,
    daily_partitioned_config,
    daily_schedule,
    default_executors,
    executor,
    failure_hook,
    graph,
    hourly_partitioned_config,
    hourly_schedule,
    in_process_executor,
    intermediate_storage,
    lambda_solid,
    logger,
    make_values_resource,
    monthly_partitioned_config,
    monthly_schedule,
    multiple_process_executor_requirements,
    multiprocess_executor,
    op,
    pipeline,
    pipeline_failure_sensor,
    reconstructable,
    repository,
    resource,
    run_status_sensor,
    schedule,
    schedule_from_partitions,
    sensor,
    solid,
    success_hook,
    weekly_partitioned_config,
    weekly_schedule,
)
from dagster.core.definitions.configurable import configured
from dagster.core.definitions.policy import Backoff, Jitter, RetryPolicy
from dagster.core.definitions.schedule import build_schedule_context
from dagster.core.definitions.sensor import build_sensor_context
from dagster.core.definitions.version_strategy import VersionStrategy
from dagster.core.errors import (
    DagsterConfigMappingFunctionError,
    DagsterError,
    DagsterEventLogInvalidForRun,
    DagsterExecutionStepExecutionError,
    DagsterExecutionStepNotFoundError,
    DagsterInvalidConfigDefinitionError,
    DagsterInvalidConfigError,
    DagsterInvalidDefinitionError,
    DagsterInvariantViolationError,
    DagsterResourceFunctionError,
    DagsterRunNotFoundError,
    DagsterStepOutputNotFoundError,
    DagsterSubprocessError,
    DagsterTypeCheckDidNotPass,
    DagsterTypeCheckError,
    DagsterUnknownResourceError,
    DagsterUnmetExecutorRequirementsError,
    DagsterUserCodeExecutionError,
)
from dagster.core.events import DagsterEvent, DagsterEventType
from dagster.core.execution.api import (
    execute_pipeline,
    execute_pipeline_iterator,
    reexecute_pipeline,
    reexecute_pipeline_iterator,
)
from dagster.core.execution.context.compute import SolidExecutionContext
from dagster.core.execution.context.hook import HookContext, build_hook_context
from dagster.core.execution.context.init import InitResourceContext, build_init_resource_context
from dagster.core.execution.context.input import InputContext, build_input_context
from dagster.core.execution.context.invocation import build_op_context, build_solid_context
from dagster.core.execution.context.logger import InitLoggerContext
from dagster.core.execution.context.output import OutputContext, build_output_context
from dagster.core.execution.context.system import TypeCheckContext
from dagster.core.execution.results import (
    CompositeSolidExecutionResult,
    PipelineExecutionResult,
    SolidExecutionResult,
)
from dagster.core.execution.validate_run_config import validate_run_config
from dagster.core.executor.base import Executor
from dagster.core.executor.init import InitExecutorContext
from dagster.core.instance import DagsterInstance
from dagster.core.launcher import DefaultRunLauncher
from dagster.core.log_manager import DagsterLogManager
from dagster.core.storage.event_log import (
    EventLogEntry,
    EventLogRecord,
    EventRecordsFilter,
    RunShardedEventsCursor,
)
from dagster.core.storage.file_manager import FileHandle, LocalFileHandle, local_file_manager
from dagster.core.storage.fs_io_manager import custom_path_fs_io_manager, fs_io_manager
from dagster.core.storage.init import InitIntermediateStorageContext
from dagster.core.storage.io_manager import IOManager, IOManagerDefinition, io_manager
from dagster.core.storage.mem_io_manager import mem_io_manager
from dagster.core.storage.memoizable_io_manager import MemoizableIOManager
from dagster.core.storage.pipeline_run import PipelineRun, PipelineRunStatus
from dagster.core.storage.root_input_manager import (
    RootInputManager,
    RootInputManagerDefinition,
    root_input_manager,
)
from dagster.core.storage.system_storage import (
    build_intermediate_storage_from_object_store,
    default_intermediate_storage_defs,
    fs_intermediate_storage,
    io_manager_from_intermediate_storage,
    mem_intermediate_storage,
)
from dagster.core.storage.tags import MEMOIZED_RUN_TAG
from dagster.core.types.config_schema import (
    DagsterTypeLoader,
    DagsterTypeMaterializer,
    dagster_type_loader,
    dagster_type_materializer,
)
from dagster.core.types.dagster_type import DagsterType, List, Optional, PythonObjectDagsterType
from dagster.core.types.decorator import (
    make_python_type_usable_as_dagster_type,
    usable_as_dagster_type,
)
from dagster.core.types.marshal import SerializationStrategy
from dagster.core.types.python_dict import Dict
from dagster.core.types.python_set import Set
from dagster.core.types.python_tuple import Tuple
from dagster.utils import file_relative_path
from dagster.utils.backcompat import ExperimentalWarning
from dagster.utils.partitions import (
    create_offset_partition_selector,
    date_partition_range,
    identity_partition_selector,
)
from dagster.utils.test import (
    check_dagster_type,
    execute_solid,
    execute_solid_within_pipeline,
    execute_solids_within_pipeline,
)

from .version import __version__

from dagster.config.source import BoolSource, StringSource, IntSource  # isort:skip


__all__ = [
    # Definition
    "AssetKey",
    "AssetMaterialization",
    "AssetSensorDefinition",
    "CompositeSolidDefinition",
    "ConfigMapping",
    "DependencyDefinition",
    "EventMetadata",
    "EventMetadataEntry",
    "ExecutorDefinition",
    "ExecutorRequirement",
    "ExpectationResult",
    "Failure",
    "Field",
    "GraphDefinition",
    "HookDefinition",
    "In",
    "InputDefinition",
    "InputMapping",
    "IntermediateStorageDefinition",
    "JsonMetadataEntryData",
    "LoggerDefinition",
    "build_init_logger_context",
    "MarkdownMetadataEntryData",
    "IntMetadataEntryData",
    "FloatMetadataEntryData",
    "Materialization",
    "ModeDefinition",
    "MultiDependencyDefinition",
    "Out",
    "Output",
    "OutputDefinition",
    "OutputMapping",
    "PathMetadataEntryData",
    "PipelineDefinition",
    "PresetDefinition",
    "PythonArtifactMetadataEntryData",
    "RepositoryData",
    "RepositoryDefinition",
    "ResourceDefinition",
    "SolidDefinition",
    "SolidInvocation",
    "TextMetadataEntryData",
    "UrlMetadataEntryData",
    "make_values_resource",
    "RetryPolicy",
    "Backoff",
    "Jitter",
    "RunStatusSensorDefinition",
    "DynamicOutput",
    "DynamicOut",
    "DynamicOutputDefinition",
    # Decorators
    "asset_sensor",
    "composite_solid",
    "executor",
    "graph",
    "intermediate_storage",
    "lambda_solid",
    "logger",
    "op",
    "pipeline",
    "repository",
    "resource",
    "schedule",
    "sensor",
    "solid",
    "success_hook",
    "failure_hook",
    "pipeline_failure_sensor",
    "run_status_sensor",
    # Execution
    "CompositeSolidExecutionResult",
    "DagsterEvent",
    "DagsterEventType",
    "DefaultRunLauncher",
    "EventLogEntry",
    "EventLogRecord",
    "Executor",
    "InitExecutorContext",
    "InitLoggerContext",
    "InitResourceContext",
    "build_init_resource_context",
    "InitIntermediateStorageContext",
    "PipelineExecutionResult",
    "RetryRequested",
    "SolidExecutionResult",
    "SolidExecutionContext",
    "build_solid_context",
    "build_op_context",
    "HookContext",
    "build_hook_context",
    "TypeCheckContext",
    "InputContext",
    "build_input_context",
    "OutputContext",
    "build_output_context",
    "PipelineRun",
    "PipelineRunStatus",
    "default_executors",
    "default_intermediate_storage_defs",
    "execute_pipeline_iterator",
    "execute_pipeline",
    "validate_run_config",
    "execute_solid_within_pipeline",
    "fs_intermediate_storage",
    "in_process_executor",
    "mem_intermediate_storage",
    "io_manager_from_intermediate_storage",
    "multiprocess_executor",
    "multiple_process_executor_requirements",
    "reconstructable",
    "reexecute_pipeline_iterator",
    "reexecute_pipeline",
    # Errors
    "DagsterConfigMappingFunctionError",
    "DagsterError",
    "DagsterEventLogInvalidForRun",
    "DagsterExecutionStepExecutionError",
    "DagsterExecutionStepNotFoundError",
    "DagsterInvalidConfigDefinitionError",
    "DagsterInvalidConfigError",
    "DagsterInvalidDefinitionError",
    "DagsterInvariantViolationError",
    "DagsterResourceFunctionError",
    "DagsterRunNotFoundError",
    "DagsterStepOutputNotFoundError",
    "DagsterSubprocessError",
    "DagsterTypeCheckDidNotPass",
    "DagsterTypeCheckError",
    "DagsterUnknownResourceError",
    "DagsterUnmetExecutorRequirementsError",
    "DagsterUserCodeExecutionError",
    # Logging
    "DagsterLogManager",
    # Utilities
    "build_intermediate_storage_from_object_store",
    "check_dagster_type",
    "execute_solid",
    "execute_solids_within_pipeline",
    "file_relative_path",
    "configured",
    # types
    "Any",
    "Bool",
    "Dict",
    "Enum",
    "EnumValue",
    "Float",
    "Int",
    "List",
    "Nothing",
    "Optional",
    "SerializationStrategy",
    "Set",
    "String",
    "Tuple",
    "TypeCheck",
    "dagster_type_loader",
    "DagsterTypeLoader",
    "dagster_type_materializer",
    "DagsterTypeMaterializer",
    # type creation
    "DagsterType",
    "PythonObjectDagsterType",
    "make_python_type_usable_as_dagster_type",
    "usable_as_dagster_type",
    # config
    "Array",
    "BoolSource",
    "ConfigSchema",
    "Noneable",
    "Permissive",
    "ScalarUnion",
    "StringSource",
    "IntSource",
    "Selector",
    "Shape",
    # file things
    "FileHandle",
    "LocalFileHandle",
    "local_file_manager",
    # instance
    "DagsterInstance",
    # storage
    "EventRecordsFilter",
    "RunShardedEventsCursor",
    # partitions and schedules
    "schedule_from_partitions",
    "daily_partitioned_config",
    "hourly_partitioned_config",
    "monthly_partitioned_config",
    "weekly_partitioned_config",
    "Partition",
    "PartitionScheduleDefinition",
    "PartitionSetDefinition",
    "RunRequest",
    "ScheduleDefinition",
    "ScheduleEvaluationContext",
    "ScheduleExecutionContext",
    "build_schedule_context",
    "SensorDefinition",
    "SensorEvaluationContext",
    "SensorExecutionContext",
    "PipelineFailureSensorContext",
    "RunStatusSensorContext",
    "build_sensor_context",
    "SkipReason",
    "daily_schedule",
    "hourly_schedule",
    "monthly_schedule",
    "weekly_schedule",
    "create_offset_partition_selector",
    "date_partition_range",
    "identity_partition_selector",
    # IO managers
    "IOManager",
    "IOManagerDefinition",
    "io_manager",
    "RootInputManager",
    "RootInputManagerDefinition",
    "root_input_manager",
    "fs_io_manager",
    "mem_io_manager",
    "custom_path_fs_io_manager",
    # warnings
    "ExperimentalWarning",
    # Versioning / Memoization
    "VersionStrategy",
    "MEMOIZED_RUN_TAG",
    "MemoizableIOManager",
]



from collections import namedtuple
from typing import Any, Set

from dagster import check
from dagster.serdes import DefaultNamedTupleSerializer, whitelist_for_serdes

from .config_type import ConfigScalarKind, ConfigType, ConfigTypeKind
from .field import Field


def get_recursive_type_keys(config_type_snap, config_schema_snapshot):
    check.inst_param(config_type_snap, "config_type_snap", ConfigTypeSnap)
    check.inst_param(config_schema_snapshot, "config_schema_snapshot", ConfigSchemaSnapshot)
    result_keys = set()
    for type_key in config_type_snap.get_child_type_keys():
        result_keys.add(type_key)
        for recurse_key in get_recursive_type_keys(
            config_schema_snapshot.get_config_snap(type_key), config_schema_snapshot
        ):
            result_keys.add(recurse_key)
    return result_keys


@whitelist_for_serdes
class ConfigSchemaSnapshot(namedtuple("_ConfigSchemaSnapshot", "all_config_snaps_by_key")):
    def __new__(cls, all_config_snaps_by_key):
        return super(ConfigSchemaSnapshot, cls).__new__(
            cls,
            all_config_snaps_by_key=check.dict_param(
                all_config_snaps_by_key,
                "all_config_snaps_by_key",
                key_type=str,
                value_type=ConfigTypeSnap,
            ),
        )

    @property
    def all_config_keys(self):
        return list(self.all_config_snaps_by_key.keys())

    def get_config_snap(self, key):
        check.str_param(key, "key")
        return self.all_config_snaps_by_key[key]

    def has_config_snap(self, key):
        check.str_param(key, "key")
        return key in self.all_config_snaps_by_key


class ConfigTypeSnapSerializer(DefaultNamedTupleSerializer):
    @classmethod
    def skip_when_empty(cls) -> Set[str]:
        return {"field_aliases"}  # Maintain stable snapshot ID for back-compat purposes


@whitelist_for_serdes(serializer=ConfigTypeSnapSerializer)
class ConfigTypeSnap(
    namedtuple(
        "_ConfigTypeSnap",
        "kind key given_name description "
        "type_param_keys "  # only valid for closed generics (Set, Tuple, List, Optional)
        "enum_values "  # only valid for enums
        "fields "  # only valid for dicts and selectors
        "scalar_kind "  # only valid for scalars
        "field_aliases",  # only valid for strict shapes
    )
):
    # serdes log
    # * Adding scalar_kind
    # * Adding field_aliases
    def __new__(
        cls,
        kind,
        key,
        given_name,
        description,
        type_param_keys,
        enum_values,
        fields,
        # Old version of object will not have these properties
        scalar_kind=None,
        field_aliases=None,
    ):
        return super(ConfigTypeSnap, cls).__new__(
            cls,
            kind=check.inst_param(kind, "kind", ConfigTypeKind),
            key=check.str_param(key, "key"),
            given_name=check.opt_str_param(given_name, "given_name"),
            type_param_keys=None
            if type_param_keys is None
            else check.list_param(type_param_keys, "type_param_keys", of_type=str),
            enum_values=None
            if enum_values is None
            else check.list_param(enum_values, "enum_values", of_type=ConfigEnumValueSnap),
            fields=None
            if fields is None
            else sorted(
                check.list_param(fields, "field", of_type=ConfigFieldSnap), key=lambda ct: ct.name
            ),
            description=check.opt_str_param(description, "description"),
            scalar_kind=check.opt_inst_param(scalar_kind, "scalar_kind", ConfigScalarKind),
            field_aliases=check.opt_dict_param(field_aliases, "field_aliases"),
        )

    @property
    def inner_type_key(self):
        # valid for Noneable and Array
        check.invariant(self.kind == ConfigTypeKind.NONEABLE or self.kind == ConfigTypeKind.ARRAY)
        check.invariant(len(self.type_param_keys) == 1)
        return self.type_param_keys[0]

    @property
    def scalar_type_key(self):
        check.invariant(self.kind == ConfigTypeKind.SCALAR_UNION)
        return self.type_param_keys[0]

    @property
    def non_scalar_type_key(self):
        check.invariant(self.kind == ConfigTypeKind.SCALAR_UNION)
        return self.type_param_keys[1]

    def _get_field(self, name):
        check.str_param(name, "name")
        check.invariant(ConfigTypeKind.has_fields(self.kind))

        for f in self.fields:
            if f.name == name:
                return f

        return None

    def get_field(self, name):
        field = self._get_field(name)
        if not field:
            check.failed("Field {name} not found".format(name=name))
        return field

    def has_field(self, name):
        return bool(self._get_field(name))

    @property
    def field_names(self):
        return [fs.name for fs in self.fields]

    def get_child_type_keys(self):
        if ConfigTypeKind.is_closed_generic(self.kind):
            return self.type_param_keys
        elif ConfigTypeKind.has_fields(self.kind):
            return [field.type_key for field in self.fields]
        else:
            return []

    def has_enum_value(self, value):
        check.invariant(self.kind == ConfigTypeKind.ENUM)
        for enum_value in self.enum_values:
            if enum_value.value == value:
                return True
        return False


@whitelist_for_serdes
class ConfigEnumValueSnap(namedtuple("_ConfigEnumValueSnap", "value description")):
    def __new__(cls, value, description):
        return super(ConfigEnumValueSnap, cls).__new__(
            cls,
            value=check.str_param(value, "value"),
            description=check.opt_str_param(description, "description"),
        )


@whitelist_for_serdes
class ConfigFieldSnap(
    namedtuple(
        "_ConfigFieldSnap",
        "name type_key is_required default_provided default_value_as_json_str description",
    )
):
    def __new__(
        cls, name, type_key, is_required, default_provided, default_value_as_json_str, description
    ):
        return super(ConfigFieldSnap, cls).__new__(
            cls,
            name=check.opt_str_param(name, "name"),
            type_key=check.str_param(type_key, "type_key"),
            is_required=check.bool_param(is_required, "is_required"),
            default_provided=check.bool_param(default_provided, "default_provided"),
            default_value_as_json_str=check.opt_str_param(
                default_value_as_json_str, "default_value_as_json_str"
            ),
            description=check.opt_str_param(description, "description"),
        )


def snap_from_field(name, field):
    check.str_param(name, "name")
    check.inst_param(field, "field", Field)
    return ConfigFieldSnap(
        name=name,
        type_key=field.config_type.key,
        is_required=field.is_required,
        default_provided=field.default_provided,
        default_value_as_json_str=field.default_value_as_json_str
        if field.default_provided
        else None,
        description=field.description,
    )


def snap_from_config_type(config_type):
    check.inst_param(config_type, "config_type", ConfigType)
    return ConfigTypeSnap(
        key=config_type.key,
        given_name=config_type.given_name,
        kind=config_type.kind,
        description=config_type.description,
        type_param_keys=[ct.key for ct in config_type.type_params] if config_type.type_params
        # jam scalar union types into type_param_keys
        else [config_type.scalar_type.key, config_type.non_scalar_type.key]
        if config_type.kind == ConfigTypeKind.SCALAR_UNION
        else None,
        enum_values=[
            ConfigEnumValueSnap(ev.config_value, ev.description) for ev in config_type.enum_values
        ]
        if config_type.kind == ConfigTypeKind.ENUM
        else None,
        fields=[snap_from_field(name, field) for name, field in config_type.fields.items()]
        if ConfigTypeKind.has_fields(config_type.kind)
        else None,
        scalar_kind=config_type.scalar_kind if config_type.kind == ConfigTypeKind.SCALAR else None,
        field_aliases=config_type.field_aliases
        if config_type.kind == ConfigTypeKind.STRICT_SHAPE
        else None,
    )


def minimal_config_for_type_snap(
    config_schema_snap: ConfigSchemaSnapshot, config_type_snap: ConfigTypeSnap
) -> Any:

    check.inst_param(config_schema_snap, "config_schema_snap", ConfigSchemaSnapshot)
    check.inst_param(config_type_snap, "config_type_snap", ConfigTypeSnap)

    if ConfigTypeKind.has_fields(config_type_snap.kind):
        default_dict = {}
        if ConfigTypeKind.is_selector(config_type_snap.kind):
            return "<selector>"
        for field in config_type_snap.fields:
            if not field.is_required:
                continue

            default_dict[field.name] = minimal_config_for_type_snap(
                config_schema_snap, config_schema_snap.get_config_snap(field.type_key)
            )
        return default_dict
    elif config_type_snap.kind == ConfigTypeKind.ANY:
        return "AnyType"
    elif config_type_snap.kind == ConfigTypeKind.SCALAR:
        defaults = {"String": "...", "Int": 0, "Float": 0.0, "Bool": True}

        return defaults.get(config_type_snap.given_name, "<unknown>")
    elif config_type_snap.kind == ConfigTypeKind.ARRAY:
        return []
    elif config_type_snap.kind == ConfigTypeKind.ENUM:
        # guard against the edge case that an enum is defined with zero options
        return (
            config_type_snap.enum_values[0].value if config_type_snap.enum_values else "<unknown>"
        )
    elif config_type_snap.kind == ConfigTypeKind.SCALAR_UNION:
        return minimal_config_for_type_snap(
            config_schema_snap,
            config_schema_snap.get_config_snap(config_type_snap.type_param_keys[0]),
        )
    else:
        return "<unknown>"
